# Description: Schema Document
from scaledp.utils.dataclass import map_dataclass_to_struct, register_type
from dataclasses import dataclass
from scaledp.schemas.Box import Box


@dataclass(order=True)
class Document:
    path: str
    text: str
    type: str
    bboxes: list[Box]
    exception: str = ""

    @staticmethod
    def get_schema():
        return map_dataclass_to_struct(Document)


register_type(Document, Document.get_schema)



import io
import gc
from pyspark import keyword_only

from scaledp.enums import Device
from scaledp.params import *
from scaledp.schemas.Box import Box
from scaledp.schemas.Document import Document
from scaledp.models.recognizers.BaseOcr import BaseOcr


class DocTROcr(BaseOcr, HasDevice, HasBatchSize):

    defaultParams = {
        "inputCol": "image",
        "outputCol": "text",
        "keepInputData": False,
        "scaleFactor": 1.0,
        "scoreThreshold": 0.5,
        "lang": ["eng"],
        "lineTolerance": 0,
        "keepFormatting": False,
        "partitionMap": False,
        "numPartitions": 0,
        "pageCol": "page",
        "pathCol": "path",
        "device": Device.CPU,
        "batchSize": 2,
        "propagateError": False,
    }

    @keyword_only
    def __init__(self, **kwargs):
        super(DocTROcr, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)

    @classmethod
    def call_ocr(cls, images, params):
        if int(params["device"]) == Device.CPU.value:
            device = False
        else:
            device = True
        import torch
        from doctr.io import DocumentFile
        from doctr.models import ocr_predictor

        # det_arch = 'db_mobilenet_v3_large',  # detection architecture
        # reco_arch = 'crnn_mobilenet_v3_small',  # recognition architecture

        model = ocr_predictor(pretrained=True)
        results = []
        for image, image_path in images:
            buff = io.BytesIO()
            image.save(buff, "png")
            doc = DocumentFile.from_images(buff.getvalue())
            # Analyze
            result = model(doc)

            boxes = []

            for page in result.pages:
                h, w = page.dimensions
                for block in page.blocks:
                    for line in block.lines:
                        for word in line.words:

                            xmin, ymin, xmax, ymax = [
                                tupl for tuploftupls in word.geometry for tupl in tuploftupls
                            ]
                            xmin = int(xmin * w)
                            ymin = int(ymin * h)
                            xmax = int(xmax * w)
                            ymax = int(ymax * h)

                            boxes.append(
                                Box(
                                    word.value,
                                    word.confidence,
                                    xmin,
                                    ymin,
                                    abs(xmax - xmin),
                                    abs(ymax - ymin),
                                )
                            )

            if params["keepFormatting"]:
                text = DocTROcr.box_to_formatted_text(boxes, params["lineTolerance"])
            else:
                text = result.render()

            results.append(Document(path=image_path, text=text, type="text", bboxes=boxes))

        gc.collect()

        if int(params["device"]) == Device.CUDA.value:
            torch.cuda.empty_cache()

        return results



import json
import time
import base64
import logging

from .BaseVisualExtractor import BaseVisualExtractor
from pyspark import keyword_only

from ...params import HasLLM, HasSchema, HasPrompt
from ...schemas.ExtractorOutput import ExtractorOutput
from tenacity import (
    retry,
    stop_after_attempt,
    wait_random_exponential,
    retry_if_exception_type
)

class LLMVisualExtractor(BaseVisualExtractor, HasLLM, HasSchema, HasPrompt):

    defaultParams = {
        "inputCol": "image",
        "outputCol": "data",
        "keepInputData": True,
        "model": "gemini-1.5-flash",
        "apiBase": None,
        "apiKey": None,
        "numPartitions": 1,
        "pageCol": "page",
        "pathCol": "path",
        "prompt": """Please extract data from the scanned image as json. Date format is yyyy-mm-dd""",
        "systemPrompt": "You are data extractor from the scanned images.",
        "delay": 30,
        "maxRetry": 6,
        "propagateError": False,
        "temperature": 1.0,
        "schemaByPrompt": True
    }

    @keyword_only
    def __init__(self, **kwargs):
        super(LLMVisualExtractor, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)
        self.pipeline = None

    def call_extractor(self, images, params):
        from openai import RateLimitError

        client = self.getOIClient()
        results = []


        @retry(retry=retry_if_exception_type(RateLimitError),
               wait=wait_random_exponential(min=1,
                                            max=self.getDelay()),
                                            stop=stop_after_attempt(self.getMaxRetry()))
        def completion_with_backoff(**kwargs):
            logging.info(f"Calling LLM API")
            return client.beta.chat.completions.parse(**kwargs)

        kwargs = {}

        content = [
            {
              "type": "text",
              "text": params["prompt"]
            }
        ]
        if self.getSchemaByPrompt():
            content.append({
                "type": "text",
                "text": "Schema for the output json: " + self.getSchema() + " Always return valid json. Do not include schema to the output."
            })
            #kwargs["response_format"] = {"type": "json_object"}
        else:
            kwargs["response_format"] = self.getPaydanticSchema()

        for image in images:
            image_decoded = base64.b64encode(image.data).decode('utf-8')
            completion = completion_with_backoff(
                model=params["model"],
                messages=[
                    {
                        "role": "system",
                        "content": params["systemPrompt"],
                        "role": "user",
                        "content": content + [
                            {
                              "type": "image_url",
                              "image_url": {
                                "url": f"data:image/jpeg;base64,{image_decoded}"
                                }
                            }
                        ]
                    },
                ],
                temperature=self.getTemperature(),
                **kwargs
            )

            result = completion.choices[0].message.content.replace("```json", "").replace("```", "")

            results.append(
                ExtractorOutput(
                    path=image.path,
                    #data=json.dumps(completion.choices[0].message.parsed.json(), indent=4, ensure_ascii=False),
                    #data=completion.choices[0].message.parsed.json(),
                    data=json.dumps(json.loads(result), indent=4, ensure_ascii=False),
                    type="LLMVisualExtractor",
                    exception="",
                )
            )
        return results



from typing import Any

import itertools


class UserDefinedFunction:
    """
    User defined function in Python

    .. versionadded:: 1.3

    Notes
    -----
    The constructor of this class is not supposed to be directly called.
    Use :meth:`pyspark.sql.functions.udf` or :meth:`pyspark.sql.functions.pandas_udf`
    to create this instance.
    """

    def __init__(
        self,
        func,
        returnType,
        name,
        evalType,
        deterministic: bool = True,
    ):
        if not callable(func):
            raise
        self.func = func
        self.returnType = returnType

    def __call__(self, *cols):
        cols = zip(*cols)
        return [self.func(*i) for i in cols]

    def _wrapped(self):
        return self


def lit(value):
    return itertools.repeat(value)


def _invoke_function(name: str, *args: Any):
    if name == "lit":
        return lit(*args)
    else:
        raise ValueError("Invalid function name: %s" % name)


temp_functions = {}


def pathSparkFunctions(pyspark):
    temp_functions["udf"] = pyspark.sql.udf.UserDefinedFunction
    pyspark.sql.udf.UserDefinedFunction = UserDefinedFunction
    temp_functions["invoke_function"] = pyspark.sql.functions._invoke_function
    pyspark.sql.functions._invoke_function = _invoke_function


def unpathSparkFunctions(pyspark):
    pyspark.sql.udf.UserDefinedFunction = temp_functions["udf"]
    pyspark.sql.functions._invoke_function = temp_functions["invoke_function"]


import pandas as pd


class DatasetPd(pd.DataFrame):

    def withColumn(self, name, col):
        self.insert(0, name, col, True)
        return self

    def drop(self, col):
        return self

    def repartition(self, numPartitions):
        return self

    def coalesce(self, numPartitions):
        return self


class PandasPipeline:
    stages = []

    def setStages(self, value):
        self.stages = value
        return self

    def __init__(self, stages):
        self.setStages(stages)

    def fromFile(self, filename):
        with open(filename, "rb") as f:
            data = f.read()

        # input = Dataset(data = [dict(content=data, path=filename)])

        input = DatasetPd(dict(content=[data], path=[filename], resolution=[0]))

        data = input
        for stage in self.stages:
            data = stage._transform(data)

        return data

    def fromBinary(self, data, filename="memory"):
        input = DatasetPd(dict(content=[data], path=[filename], resolution=[0]))
        data = input
        for stage in self.stages:
            data = stage._transform(data)

        return data



import json

from .BaseVisualExtractor import BaseVisualExtractor
from pyspark import keyword_only

from ...params import HasLLM, HasSchema, HasPrompt
from ...schemas.ExtractorOutput import ExtractorOutput
from ...utils.pydantic_shema_utils import json_schema_to_model


class GeminiVisualExtractor(BaseVisualExtractor, HasLLM, HasSchema, HasPrompt):

    defaultParams = {
        "inputCol": "image",
        "outputCol": "data",
        "keepInputData": True,
        "model": "gemini-1.5-flash",
        "apiBase": "",
        "apiKey": "",
        "numPartitions": 1,
        "pageCol": "page",
        "pathCol": "path",
        "prompt": """Please extract data from the scanned image as json.""",
    }

    @keyword_only
    def __init__(self, **kwargs):
        super(GeminiVisualExtractor, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)
        self.pipeline = None

    def call_extractor(self, images, params):
        import google.generativeai as genai
        import base64
        import json

        schema = json.loads(params["schema"])
        schema = json_schema_to_model(schema, schema.get("$defs", {}))
        generation_config = {
            "temperature": 1,
            "top_p": 0.95,
            "top_k": 40,
            "max_output_tokens": 8192,
            "response_mime_type": "application/json",
            "response_schema": schema
        }

        genai.configure(api_key=params["apiKey"])

        model = genai.GenerativeModel(
            model_name=params["model"],
            generation_config=generation_config,
        )

        schema = json.loads(params["schema"])
        schema = json_schema_to_model(schema, schema.get("$defs", {}))

        results = []

        for image in images:
            image_decoded = base64.b64encode(image.data).decode('utf-8')
            prompt = params["prompt"].format(schema=schema)
            response = model.generate_content([{'mime_type':'image/png', 'data': image_decoded}, prompt], stream=False)
            results.append(
                ExtractorOutput(
                    path=image.path,
                    data=response.text,
                    type="GeminiVisualExtractor",
                    exception="",
                )
            )
        return results



from pyspark import keyword_only

from ...enums import Device
from scaledp.params import *
from scaledp.schemas.Box import Box
from scaledp.schemas.Document import Document
from scaledp.models.recognizers.BaseOcr import BaseOcr
import gc


class SuryaOcr(BaseOcr, HasDevice, HasBatchSize):

    defaultParams = {
        "inputCol": "image",
        "outputCol": "text",
        "keepInputData": False,
        "scaleFactor": 1.0,
        "scoreThreshold": 0.5,
        "lang": ["eng"],
        "lineTolerance": 0,
        "keepFormatting": False,
        "partitionMap": False,
        "numPartitions": 0,
        "pageCol": "page",
        "pathCol": "path",
        "device": Device.CPU,
        "batchSize": 2,
        "propagateError": False,
    }

    @keyword_only
    def __init__(self, **kwargs):
        super(SuryaOcr, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)

    @classmethod
    def call_ocr(cls, images, params):
        import torch
        from surya.ocr import run_ocr
        from surya.model.detection.model import (
            load_model as load_det_model,
            load_processor as load_det_processor,
        )
        from surya.model.recognition.model import load_model as load_rec_model
        from surya.model.recognition.processor import (
            load_processor as load_rec_processor,
        )
        from surya.settings import settings

        if int(params["device"]) == Device.CPU.value:
            device = "cpu"
        else:
            device = "cuda"

        langs = params["lang"]

        settings.DETECTOR_BATCH_SIZE = params["batchSize"]
        settings.RECOGNITION_BATCH_SIZE = params["batchSize"]

        det_processor, det_model = load_det_processor(), load_det_model(device=device)
        rec_model, rec_processor = load_rec_model(device=device), load_rec_processor()

        results = []
        predictions = run_ocr(
            [image[0] for image in images],
            [langs] * len(images),
            det_model,
            det_processor,
            rec_model,
            rec_processor,
        )

        for prediction, (image, image_path) in zip(predictions, images):
            boxes = [
                Box(
                    text=x.text,
                    score=x.confidence,
                    x=x.bbox[0],
                    y=x.bbox[1],
                    width=x.bbox[2] - x.bbox[0],
                    height=x.bbox[3] - x.bbox[1],
                )
                .toString()
                .scale(1 / params["scaleFactor"])
                for x in prediction.text_lines
            ]

            if params["keepFormatting"]:
                text = SuryaOcr.box_to_formatted_text(boxes, params["lineTolerance"])
            else:
                text = "\n".join([str(w.text) for w in boxes])

            results.append(Document(path=image_path, text=text, type="text", bboxes=boxes))

        gc.collect()
        if int(params["device"]) == Device.CUDA.value:
            torch.cuda.empty_cache()

        return results



import statistics


def cluster(items, maxgap, key=lambda x: x):
    """Arrange data into groups where successive elements
    differ by no more than *maxgap*
    """
    if not items:
        return []
    items.sort(key=key)
    groups = [[items[0]]]
    for x in items[1:]:
        if abs(key(x) - key(groups[-1][-1])) <= maxgap:
            groups[-1].append(x)
        else:
            groups.append([x])
    return groups


def get_size(items, key=lambda x: x):
    if not items:
        return 0
    items = [key(x) for x in items]
    if len(items) == 1:
        return items[0]
    items.sort()
    if len(items) <= 4:
        return int(statistics.mean(items))
    else:
        l = int(len(items) / 4)
        return int(statistics.mode(items[l:-l]))



from pyspark import keyword_only

from ...enums import Device
from scaledp.models.detectors.BaseDetector import BaseDetector
from scaledp.params import HasDevice, HasBatchSize
import gc
import os.path
from huggingface_hub import hf_hub_download

from scaledp.schemas.Box import Box
from scaledp.schemas.DetectorOutput import DetectorOutput


class YoloDetector(BaseDetector, HasDevice, HasBatchSize):
    _model = None

    defaultParams = {
        "inputCol": "image",
        "outputCol": "boxes",
        "keepInputData": False,
        "scaleFactor": 1.0,
        "scoreThreshold": 0.5,
        "device": Device.CPU,
        "batchSize": 2,
        "partitionMap": False,
        "numPartitions": 0,
        "pageCol": "page",
        "pathCol": "path",
        "propagateError": False,
    }

    @keyword_only
    def __init__(self, **kwargs):
        super(YoloDetector, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)
        self.get_model({k.name: v for k, v in self.extractParamMap().items()})

    @classmethod
    def get_model(cls, params):
        if cls._model:
            return cls._model
        from ultralytics import YOLO
        model = params["model"]
        if not os.path.isfile(model):
            model = hf_hub_download(repo_id=model, filename="best.pt")

        detector = YOLO(model)

        if int(params["device"]) == Device.CPU.value:
            device = "cpu"
        else:
            device = "cuda"
        cls._model = detector.to(device)
        return cls._model

    @classmethod
    def call_detector(cls, images, params):
        import torch

        detector = cls.get_model(params)

        results = detector(
            [image[0] for image in images],
            conf=params["scoreThreshold"],
            save_conf=True,
        )

        results_final = []
        for res, (image, image_path) in zip(results, images):
            boxes = []
            for box in res.boxes:
                boxes.append(Box.fromBBox(box.xyxy[0]))
            results_final.append(DetectorOutput(path=image_path, type="yolo", bboxes=boxes))

        gc.collect()
        if int(params["device"]) == Device.CUDA.value:
            torch.cuda.empty_cache()

        return results_final



from pyspark.sql import DataFrame
from scaledp import DataToImage, PdfDataToImage, TesseractOcr, TextToDocument
import pyspark.sql.functions as f
import random
import base64
import json
import logging


def _show_image(image, width=600, show_meta=True, index=0):
    from IPython.display import display, HTML
    from jinja2 import PackageLoader, Environment

    if image is None:
        print("Empty image")
        return

    img_base64 = base64.b64encode(image.data).decode("utf-8")

    templateEnv = Environment(loader=PackageLoader("scaledp.utils", "templates"))
    template = templateEnv.get_template("image.html")
    metadata = {
        "Image#": index,
        "Path": image.path.split("/")[-1],
        "Size": f"{image.width} x {image.height} px",
        "Resolution": f"{image.resolution} dpi",
    }

    if image.exception != "":
        metadata["Exception"] = image.exception

    rendered_html = template.render(width=width, metadata=metadata, image=img_base64)

    display(HTML(rendered_html))


def get_column_type(df: DataFrame, column_name: str) -> str:
    for name, dtype in df.dtypes:
        if name == column_name:
            return dtype
    return None


def show_image(df, column="", limit=5, width=600, show_meta=True):
    if column == "":
        if "image" in df.columns:
            column = "image"
        elif "content" in df.columns:
            column = "content"
        else:
            raise ValueError("Please specify column name")
    column_type = get_column_type(df, column)
    if column_type == "binary":
        df = DataToImage().setInputCol(column).setOutputCol("image").transform(df)
        column = "image"
    for id_, row in enumerate(df.limit(limit).select(column).collect()):
        image = row[column]
        _show_image(image, width, show_meta, id_)


def show_text(df, column="", field="text", limit=5, width=800):
    from IPython.display import display, HTML
    from jinja2 import PackageLoader, Environment

    templateEnv = Environment(loader=PackageLoader("scaledp.utils", "templates"))
    template = templateEnv.get_template("text.html")
    df = df.limit(limit)
    if column == "":
        if "value" in df.columns:
            column = "text"
            df = TextToDocument(inputCol="value").transform(df)
        elif "text" in df.columns:
            column = "text"
        elif "content" in df.columns:
            column = "text"
            df = DataToImage().transform(df)
            df = TesseractOcr(keepFormatting=True).transform(df)
        elif "image" in df.columns:
            column = "text"
            df = TesseractOcr(keepFormatting=True).transform(df)
        else:
            raise ValueError("Please specify column name")
    for id, text in enumerate(df.select(f"{column}.*").collect()):
        metadata = {
            "Id": id,
            "Path": text.path.split("/")[-1],
            "Exception": text.exception,
        }

        rendered_html = template.render(width=width, metadata=metadata, text=getattr(text, field))

        display(HTML(rendered_html))

def show_json(df, column="", field="data", limit=5, width=800):
    from IPython.display import display, HTML
    from jinja2 import PackageLoader, Environment

    templateEnv = Environment(loader=PackageLoader("scaledp.utils", "templates"))
    template = templateEnv.get_template("json.html")
    df = df.limit(limit)
    if column == "":
        if "data" in df.columns:
            column = "data"
        elif "text" in df.columns:
            column = "text"
        else:
            raise ValueError("Please specify column name")
    for id, text in enumerate(df.select(f"{column}.*").collect()):
        metadata = {
            "Id": id,
            "Path": text.path.split("/")[-1],
            "Exception": text.exception,
        }
        data = getattr(text, field)
        try:
            data = json.dumps(json.loads(data), indent=4, ensure_ascii=False)
        except:
            logging.warning("Failed to parse json")

        from IPython.display import Code
        data = Code(data, language='json')._repr_html_()
        rendered_html = template.render(width=width, metadata=metadata, text=data)

        display(HTML(rendered_html))

def show_pdf(df, column="", limit=5, width=600, show_meta=True):
    if column == "":
        if "pdf" in df.columns:
            column = "pdf"
        elif "content" in df.columns:
            column = "content"
        else:
            raise ValueError("Please specify column name")
    column_type = get_column_type(df, column)
    if column_type == "binary":
        df = PdfDataToImage(inputCol=column).transform(df)
        column = "image"
    else:
        raise ValueError("Column must be binary")
    for id_, row in enumerate(df.limit(limit).select(column).collect()):
        image = row[column]
        _show_image(image, width, show_meta, id_)


def show_ner(df, column="ner", limit=20, truncate=False):
    df.select(f.explode(f"{column}.entities").alias("ner")).select("ner.*").show(
        limit, truncate=truncate
    )


def visualize_ner(df, column="ner", text_column="text", limit=20, width=800, labels_list=None):
    from IPython.display import display, HTML
    from jinja2 import PackageLoader, Environment

    templateEnv = Environment(loader=PackageLoader("scaledp.utils", "templates"))
    template = templateEnv.get_template("ner.html")

    df = df.limit(limit).select(column, text_column).cache()
    entities = df.select(f.explode(f"{column}.entities").alias("ner")).select("ner.*").collect()
    text = df.select(text_column).collect()[0][0]
    original_text = text.text

    entity_colors = {}
    current_position = 0
    html = ""
    for entity in entities:
        start = entity.start
        end = entity.end
        entity_name = entity.entity_group.lower()
        if entity_name not in entity_colors:
            entity_colors[entity_name] = "#{:06x}".format(random.randint(0, 0xFFFFFF))

        if current_position < start:
            html += (
                "<span style='font-size:16px;line-height: 25px;'>"
                + original_text[current_position:start]
                + "</span>"
            )
        if entity_name in entity_colors:
            html += f"""<span style='border-radius:4px;padding:2px;color:white;line-height:25px;margin:1px;
            background-color:{entity_colors[entity_name]};font-size:16px;white-space: nowrap;'>
            <span style="color:black;background-color:white;border-radius: 2px;padding: 0 1px 0 1px;">
            {original_text[start:end]}</span><span style='font-weight:500;padding: 4px;'>{entity.entity_group}</span></span>"""
        else:
            html += (
                "<span style='font-size:16px;line-height: 25px;'>"
                + original_text[start:end]
                + "</span>"
            )
        current_position = end

    metadata = {"Id": 0, "Path": text.path.split("/")[-1], "Exception": text.exception}

    rendered_html = template.render(width=width, metadata=metadata, ner=html)

    display(HTML(rendered_html))



import traceback
import logging
import fitz
from pyspark import keyword_only
from pyspark.sql.functions import udf, posexplode_outer
from pyspark.sql.types import ArrayType

from scaledp.schemas.Image import Image
from pyspark.ml import Transformer
from pyspark.ml.util import DefaultParamsReadable, DefaultParamsWritable
from scaledp.params import *
from scaledp.enums import ImageType


class PdfDataToImage(
    Transformer,
    HasInputCol,
    HasOutputCol,
    HasKeepInputData,
    HasImageType,
    HasPathCol,
    HasResolution,
    HasPageCol,
    DefaultParamsReadable,
    DefaultParamsWritable,
    HasColumnValidator,
    HasDefaultEnum,
):
    """
    Extract image from PDF file
    """

    pageLimit = Param(
        Params._dummy(),
        "pageLimit",
        "Limit number of pages to convert to image",
        typeConverter=TypeConverters.toInt,
    )

    defaultParams = {
        "inputCol": "content",
        "outputCol": "image",
        "pathCol": "path",
        "pageCol": "page",
        "keepInputData": False,
        "imageType": ImageType.FILE,
        "resolution": 300,
        "pageLimit": 0,
    }

    @keyword_only
    def __init__(self, **kwargs):
        super(PdfDataToImage, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)

    def transform_udf(self, input, path):
        logging.info("Run Pdf Data to Image")
        try:
            doc = fitz.open("pdf", input)
            if len(doc) == 0:
                raise ValueError("Empty PDF document.")
            if self.getPageLimit():
                doc = doc[: self.getPageLimit()]
            for page in doc:
                pix = page.get_pixmap(
                    matrix=fitz.Identity,
                    dpi=self.getResolution(),
                    colorspace=fitz.csRGB,
                    clip=None,
                    alpha=False,
                    annots=True,
                )
                yield Image.from_binary(
                    pix.pil_tobytes("png"),
                    path,
                    self.getImageType(),
                    resolution=self.getResolution(),
                    width=pix.width,
                    height=pix.height,
                )

        except Exception:
            exception = traceback.format_exc()
            exception = (
                f"{self.uid}: Error during extract image from the PDF document: {exception}"
            )
            logging.warning(exception)
            return [Image(path=path, exception=exception)]

    def _transform(self, dataset):
        out_col = self.getOutputCol()
        input_col = self._validate(self.getInputCol(), dataset)
        path_col = dataset[self.getPathCol()]

        selCol = dataset.columns + [
            posexplode_outer(
                udf(self.transform_udf, ArrayType(Image.get_schema()))(input_col, path_col)
            ).alias(self.getPageCol(), out_col),
        ]

        result = dataset.select(selCol)
        if not self.getKeepInputData():
            result = result.drop(input_col)
        return result

    def getPageLimit(self):
        return self.getOrDefault(self.pageLimit)

    def setPageLimit(self, value):
        return self._set(pageNumber=value)



import gc
from pyspark import keyword_only
import numpy as np

from scaledp.enums import Device
from scaledp.params import *
from scaledp.schemas.Box import Box
from scaledp.schemas.Document import Document
from scaledp.models.recognizers.BaseOcr import BaseOcr


class EasyOcr(BaseOcr, HasDevice, HasBatchSize):

    defaultParams = {
        "inputCol": "image",
        "outputCol": "text",
        "keepInputData": False,
        "scaleFactor": 1.0,
        "scoreThreshold": 0.5,
        "lang": ["eng"],
        "lineTolerance": 0,
        "keepFormatting": False,
        "partitionMap": False,
        "numPartitions": 0,
        "pageCol": "page",
        "pathCol": "path",
        "device": Device.CPU,
        "batchSize": 2,
        "propagateError": False,
    }

    @keyword_only
    def __init__(self, **kwargs):
        super(EasyOcr, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)

    @staticmethod
    def points_to_box(points, text, score):
        """
        Convert a set of four corner points to (x, y, width, height).

        Args:
            points (list of tuple): List of four (x, y) tuples representing the corners.

        Returns:
            tuple: A tuple (x, y, width, height).
        """
        x_coords = [p[0] for p in points]
        y_coords = [p[1] for p in points]

        x = min(x_coords)
        y = min(y_coords)
        width = max(x_coords) - x
        height = max(y_coords) - y

        return Box(text=text, score=score, x=x, y=y, width=width, height=height)

    @classmethod
    def call_ocr(cls, images, params):
        import easyocr
        import torch

        if int(params["device"]) == Device.CPU.value:
            device = False
        else:
            device = True

        langs = params["lang"]
        scale_factor = params["scaleFactor"]
        reader = easyocr.Reader(langs, device)
        results = []
        for image, image_path in images:

            image = np.array(image.convert("RGB"))[:, :, ::-1].copy()
            result = reader.readtext(image)
            boxes = [
                EasyOcr.points_to_box(box, text, float(score)).toString().scale(1 / scale_factor)
                for box, text, score in result
            ]

            if params["keepFormatting"]:
                text = EasyOcr.box_to_formatted_text(boxes, params["lineTolerance"])
            else:
                text = "\n".join([str(w.text) for w in boxes])

            results.append(Document(path=image_path, text=text, type="text", bboxes=boxes))

        gc.collect()
        if int(params["device"]) == Device.CUDA.value:
            torch.cuda.empty_cache()

        return results



import io
import logging
import base64
from pyspark import keyword_only

from scaledp.params import *
from scaledp.schemas.Box import Box
from scaledp.schemas.Document import Document
from scaledp.models.recognizers.BaseOcr import BaseOcr
from tenacity import (
    retry,
    stop_after_attempt,
    wait_random_exponential,
    retry_if_exception_type
)


class LLMOcr(BaseOcr, HasLLM, HasPrompt):

    defaultParams = {
        "inputCol": "image",
        "outputCol": "text",
        "keepInputData": False,
        "scaleFactor": 1.0,
        "scoreThreshold": 0.5,
        "lang": ["eng"],
        "lineTolerance": 0,
        "keepFormatting": False,
        "partitionMap": False,
        "numPartitions": 0,
        "pageCol": "page",
        "pathCol": "path",
        "systemPrompt": "You are ocr.",
        "prompt": """Please extract text from the image.""",
        "model": "gemini-1.5-flash",
        "apiBase": None,
        "apiKey": None,
        "delay": 30,
        "maxRetry": 6,
        "propagateError": False,
    }

    @keyword_only
    def __init__(self, **kwargs):
        super(LLMOcr, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)

    @classmethod
    def call_ocr(cls, images, params):
        from openai import RateLimitError

        client = cls.getClient(params["apiKey"], params["apiBase"])
        results = []

        @retry(retry=retry_if_exception_type(RateLimitError),
               wait=wait_random_exponential(min=1,
                                            max=params["delay"]),
               stop=stop_after_attempt(params["maxRetry"]))
        def completion_with_backoff(**kwargs):
            logging.info(f"Calling LLM API")
            return client.beta.chat.completions.parse(**kwargs)

        for image, image_path in images:
            buff = io.BytesIO()
            image.save(buff, "png")
            image_decoded = base64.b64encode(buff.getvalue()).decode('utf-8')
            completion = completion_with_backoff(
                model=params["model"],
                messages=[
                    {
                        "role": "system",
                        "content": params["systemPrompt"],
                        "role": "user",
                        "content": [
                            {
                                "type": "image_url",
                                "image_url": {
                                    "url": f"data:image/jpeg;base64,{image_decoded}"
                                }
                            },
                            {
                                "type": "text",
                                "text": params["prompt"]
                            }
                        ]
                    },
                ],
                #response_format=self.getPaydanticSchema(),
            )

            results.append(
                Document(path=image_path, text=completion.choices[0].message.content, type="text", bboxes=[])
            )
        return results



from scaledp.models.recognizers.BaseOcr import BaseOcr
from scaledp.params import HasInputCols
from pyspark.sql.functions import udf, pandas_udf, lit
import pandas as pd
from scaledp.schemas.DetectorOutput import DetectorOutput
from scaledp.schemas.Document import Document
from scaledp.schemas.Image import Image
import logging
import json
import traceback


class BaseRecognizer(BaseOcr, HasInputCols):

    def transform_udf(self, image, boxes, params=None):
        logging.info("Run Text Recognizer")
        if params is None:
            params = self.get_params()
        params = json.loads(params)
        if not isinstance(image, Image):
            image = Image(**image.asDict())

        if not isinstance(boxes, DetectorOutput):
            boxes = DetectorOutput(**boxes.asDict())
        if image.exception != "":
            return Document(
                path=image.path,
                text="",
                bboxes=[],
                type="text",
                exception=image.exception,
            )
        try:
            image_pil = image.to_pil()
            scale_factor = self.getScaleFactor()
            if scale_factor != 1.0:
                resized_image = image_pil.resize(
                    (
                        int(image_pil.width * scale_factor),
                        int(image_pil.height * scale_factor),
                    )
                )
            else:
                resized_image = image_pil

            result = self.call_recognizer([(resized_image, image.path)], [boxes], params)
        except Exception:
            exception = traceback.format_exc()
            exception = f"{self.uid}: Error in text recognition: {exception}, {image.exception}"
            logging.warning(f"{self.uid}: Error in text recognition.")
            return Document(path=image.path, text="", bboxes=[], type="ocr", exception=exception)
        return result[0]

    @classmethod
    def transform_udf_pandas(
        cls, images: pd.DataFrame, boxes: pd.DataFrame, params: pd.Series
    ) -> pd.DataFrame:
        params = json.loads(params[0])
        resized_images = []
        boxes_o = []
        for (index, image), (i, box) in zip(images.iterrows(), boxes.iterrows()):
            if not isinstance(image, Image):
                image = Image(**image.to_dict())
            if not isinstance(box, DetectorOutput):
                box = DetectorOutput(**box.to_dict())
            image_pil = image.to_pil()
            scale_factor = params["scaleFactor"]
            if scale_factor != 1.0:
                resized_image = image_pil.resize(
                    (
                        int(image_pil.width * scale_factor),
                        int(image_pil.height * scale_factor),
                    )
                )
            else:
                resized_image = image_pil
            resized_images.append((resized_image, image.path))
            boxes_o.append(box)

        results = cls.call_recognizer(resized_images, boxes_o, params)

        return pd.DataFrame(results)

    @classmethod
    def call_recognizer(cls, resized_images, boxes, params):
        raise NotImplementedError("Subclasses should implement this method")

    def _transform(self, dataset):
        out_col = self.getOutputCol()
        image_col = self._validate(self.getInputCols()[0], dataset)
        box_col = self._validate(self.getInputCols()[1], dataset)
        params = self.get_params()

        if not self.getPartitionMap():
            result = dataset.withColumn(
                out_col,
                udf(self.transform_udf, Document.get_schema())(image_col, box_col, lit(params)),
            )
        else:
            if self.getNumPartitions() > 0:
                if self.getPageCol() in dataset.columns:
                    dataset = dataset.repartition(self.getPageCol())
                elif self.getPathCol() in dataset.columns:
                    dataset = dataset.repartition(self.getPathCol())
                dataset = dataset.coalesce(self.getNumPartitions())
            result = dataset.withColumn(
                out_col,
                pandas_udf(self.transform_udf_pandas, self.outputSchema())(
                    image_col, box_col, lit(params)
                ),
            )

        if not self.getKeepInputData():
            result = result.drop(image_col)
        return result



import os
import sys
import pyspark
from pyspark import SparkConf
from pyspark.sql import SparkSession
from pyspark.ml.pipeline import PipelineModel
from importlib.resources import files
from scaledp.image.DataToImage import DataToImage
from scaledp.pdf.PdfDataToImage import PdfDataToImage
from scaledp.models.recognizers.TesseractOcr import TesseractOcr
from scaledp.models.ner.Ner import Ner
from scaledp.image.ImageDrawBoxes import ImageDrawBoxes
from scaledp.image.ImageCropBoxes import ImageCropBoxes
from scaledp.text.TextToDocument import TextToDocument
from scaledp.models.recognizers.SuryaOcr import SuryaOcr
from scaledp.models.recognizers.EasyOcr import EasyOcr
from scaledp.models.recognizers.DocTROcr import DocTROcr
from scaledp.models.detectors.YoloDetector import YoloDetector
from scaledp.models.extractors.DSPyExtractor import DSPyExtractor
from scaledp.models.recognizers.TesseractRecognizer import TesseractRecognizer
from scaledp.models.detectors.DocTRTextDetector import DocTRTextDetector
from scaledp.models.extractors.LLMVisualExtractor import LLMVisualExtractor
from scaledp.models.extractors.LLMExtractor import LLMExtractor
from scaledp.models.recognizers.LLMOcr import LLMOcr
from scaledp.models.ner.LLMNer import LLMNer
from importlib import resources
from scaledp import enums
from scaledp.enums import *

from pyspark.sql import DataFrame

from scaledp.utils.show_utils import (
    show_image,
    show_pdf,
    show_ner,
    visualize_ner,
    show_text,
    show_json,
)

DataFrame.show_image = lambda self, column="", limit=5, width=None, show_meta=True,: show_image(
    self, column, limit, width, show_meta
)
DataFrame.show_pdf = lambda self, column="", limit=5, width=None, show_meta=True,: show_pdf(
    self, column, limit, width, show_meta
)
DataFrame.show_ner = lambda self, column="ner", limit=20, truncate=True: show_ner(
    self, column, limit, truncate
)
DataFrame.show_text = lambda self, column="", field="text", limit=20, width=None: show_text(
    self, column, field, limit, width
)
DataFrame.show_json = lambda self, column="", field="data", limit=20, width=None: show_json(
    self, column, field, limit, width
)
DataFrame.visualize_ner = lambda self, column="ner", text_column="text", limit=20, width=None, labels_list=None: visualize_ner(
    self, column, text_column, limit, width, labels_list
)


def version():
    version_path = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(version_path, "VERSION"), encoding="utf-8") as version_file:
        return version_file.read().strip()


__version__ = version()


files = lambda path: resources.files("scaledp").joinpath(path).as_posix()


def aws_version():
    spark_hadoop_map = {
        "3.0": "2.7.4",
        "3.1": "3.2.0",
        "3.2": "3.3.1",
        "3.3": "3.3.2",
        "3.4": "3.3.4",
        "3.5": "3.3.4",
    }
    return spark_hadoop_map[pyspark.__version__[:3]]


def ScaleDPSession(
    conf=None, master_url="local[*]", with_aws=False, with_pro=False, logLevel="ERROR"
):
    """
    Start Spark session with ScaleDP
    @param conf: Instance of SparkConf or dict with extra configuration.
    @param master_url: Spark master URL
    @param with_aws: Enable AWS support
    @param logLevel: Log level
    """
    os.environ["PYSPARK_PYTHON"] = sys.executable
    os.environ["TRANSFORMERS_VERBOSITY"] = logLevel.lower()

    if with_pro:
        try:
            import scaledp_pro
        except ImportError:
            raise ImportError(
                "ScaleDP Pro is not installed. Please install it using 'pip install scaledp-pro'"
            )

    jars = []
    jars_packages = []
    default_conf = {
        "spark.serializer": "org.apache.spark.serializer.KryoSerializer",
        "spark.kryoserializer.buffer.max": "200M",
        "spark.driver.memory": "8G",
    }

    if with_aws:
        jars_packages.append("org.apache.hadoop:hadoop-aws:" + aws_version())

    if conf:
        if not isinstance(conf, dict):
            conf = dict(conf.getAll())
        default_conf.update(conf)
        extra_jars_packages = default_conf.get("spark.jars.packages")
        if extra_jars_packages:
            jars_packages.append(extra_jars_packages)
        extra_jars = default_conf.get("spark.jars")
        if extra_jars:
            jars.append(extra_jars)

    builder = SparkSession.builder.master(master_url).appName("ScaleDP: v" + __version__)

    for k, v in default_conf.items():
        builder.config(str(k), str(v))

    builder.config("spark.jars", ",".join(jars))
    builder.config("spark.jars.packages", ",".join(jars_packages))

    spark = builder.getOrCreate()
    spark.sparkContext.setLogLevel(logLevel=logLevel)
    return spark


__all__ = [
    "ScaleDPSession",
    "DataToImage",
    "ImageDrawBoxes",
    "PdfDataToImage",
    "TesseractOcr",
    "Ner",
    "TextToDocument",
    "PipelineModel",
    "SuryaOcr",
    "EasyOcr",
    "DocTROcr",
    "YoloDetector",
    "ImageCropBoxes",
    "DSPyExtractor",
    "TesseractRecognizer",
    "DocTRTextDetector",
    "LLMVisualExtractor",
    "LLMExtractor",
    "LLMOcr",
    "LLMNer",
    "__version__",
    "files",
] + dir(enums)



import json
import logging
import traceback
from pyspark.sql.types import *
from pyspark.ml import Transformer
from pyspark.ml.util import DefaultParamsReadable, DefaultParamsWritable
from pyspark.sql.functions import lit, udf

from scaledp.params import *
from scaledp.schemas.Document import Document
from scaledp.schemas.ExtractorOutput import ExtractorOutput


class ExtractorError(Exception):
    pass

class BaseExtractor(
    Transformer,
    HasInputCol,
    HasOutputCol,
    HasKeepInputData,
    HasPathCol,
    DefaultParamsReadable,
    DefaultParamsWritable,
    HasNumPartitions,
    HasPageCol,
    HasColumnValidator,
    HasDefaultEnum,
    HasPropagateError
):

    def get_params(self):
        return json.dumps({k.name: v for k, v in self.extractParamMap().items()})

    @classmethod
    def call_extractor(cls, documents, params):
        raise NotImplementedError("Subclasses should implement this method")

    def transform_udf(self, document, params=None):
        logging.info("Run Data Extractor")
        if params is None:
            params = self.get_params()
        params = json.loads(params)
        if not isinstance(document, Document):
            document = Document(**document.asDict())
        if document.exception != "":
            return ExtractorOutput(
                path=document.path,
                data="",
                type="extractor",
                exception=document.exception,
            )
        try:

            result = self.call_extractor([document], params)
        except Exception as e:
            exception = traceback.format_exc()
            exception = f"{self.uid}: Error in data extraction: {exception}, {document.exception}"
            logging.warning(f"{self.uid}: Error in data extraction.")
            if self.getPropagateError():
                raise ExtractorError() from e
            return ExtractorOutput(
                path=document.path, data="", type="detector", exception=exception
            )
        return result[0]

    def _transform(self, dataset):
        params = self.get_params()
        out_col = self.getOutputCol()
        in_col = self._validate(self.getInputCol(), dataset)

        result = dataset.withColumn(
            out_col,
            udf(self.transform_udf, ExtractorOutput.get_schema())(in_col, lit(params)),
        )
        if not self.getKeepInputData():
            result = result.drop(in_col)
        return result



import traceback
import logging
import json

import pandas as pd
from pyspark.sql.types import *
from scaledp.params import *
from pyspark.ml import Transformer
from pyspark.ml.util import DefaultParamsReadable, DefaultParamsWritable

from ...utils import get_size, cluster
from pyspark.sql.functions import udf, pandas_udf, lit
from scaledp.schemas.Document import Document
from scaledp.schemas.Image import Image


class OcrError(Exception):
    pass

class BaseOcr(
    Transformer,
    HasInputCol,
    HasOutputCol,
    HasKeepInputData,
    HasDefaultEnum,
    DefaultParamsReadable,
    DefaultParamsWritable,
    HasScoreThreshold,
    HasLang,
    HasPartitionMap,
    HasColumnValidator,
    HasNumPartitions,
    HasPageCol,
    HasPathCol,
    HasPropagateError,
):

    scaleFactor = Param(
        Params._dummy(),
        "scaleFactor",
        "Scale Factor.",
        typeConverter=TypeConverters.toFloat,
    )

    keepFormatting = Param(
        Params._dummy(),
        "keepFormatting",
        "Whether to keep the original formatting.",
        typeConverter=TypeConverters.toBoolean,
    )

    lineTolerance = Param(
        Params._dummy(),
        "lineTolerance",
        "Tolerance for line clustering.",
        typeConverter=TypeConverters.toInt,
    )

    @staticmethod
    def to_formatted_text(lines, character_height):
        output_lines = []
        space_width = BaseOcr.get_character_width(lines)
        y = 0
        for regions in lines:
            line = ""
            # Add extra empty lines if need
            line_diffs = int((regions[0].y - y) / (character_height * 2))
            y = regions[0].y
            if line_diffs > 1:
                for i in range(line_diffs - 1):
                    output_lines.append("")

            prev = 0
            for region in regions:
                left2 = region.x - prev
                spaces = max(int(left2 / space_width), 1)
                line = line + spaces * " " + region.text
                prev = region.x + region.width
            output_lines.append(line)
        return "\n".join(output_lines)

    @staticmethod
    def get_character_width(lines):
        character_widths = []
        for regions in lines:
            for region in regions:
                width = region.width
                if len(region.text) > 0:
                    character_widths.append(int(width / len(region.text)))
        return get_size(character_widths)

    @staticmethod
    def box_to_formatted_text(boxes, line_tolerance=0):
        character_height = get_size(boxes, lambda x: x.height)
        if line_tolerance == 0:
            line_tolerance = character_height / 3
        lines = cluster(boxes, line_tolerance, key=lambda i: int(i.y))

        lines = [sorted(xs, key=lambda i: int(i.x)) for xs in lines]
        return BaseOcr.to_formatted_text(lines, character_height)

    def transform_udf(self, image, params=None):
        logging.info("Run OCR")
        if params is None:
            params = self.get_params()
        params = json.loads(params)
        if not isinstance(image, Image):
            image = Image(**image.asDict())
        if image.exception != "":
            return Document(
                path=image.path,
                text="",
                bboxes=[],
                type="text",
                exception=image.exception,
            )
        try:
            image_pil = image.to_pil()
            scale_factor = self.getScaleFactor()
            if scale_factor != 1.0:
                resized_image = image_pil.resize(
                    (
                        int(image_pil.width * scale_factor),
                        int(image_pil.height * scale_factor),
                    )
                )
            else:
                resized_image = image_pil

            result = self.call_ocr([(resized_image, image.path)], params)
        except Exception as e:
            exception = traceback.format_exc()
            exception = f"{self.uid}: Error in text recognition: {exception}, {image.exception}"
            logging.warning(f"{self.uid}: Error in text recognition.")
            if self.getPropagateError():
                raise OcrError() from e
            return Document(path=image.path, text="", bboxes=[], type="ocr", exception=exception)
        return result[0]

    @classmethod
    def call_ocr(cls, resized_images, params):
        raise NotImplementedError("Subclasses should implement this method")

    @classmethod
    def transform_udf_pandas(cls, images: pd.DataFrame, params: pd.Series) -> pd.DataFrame:
        params = json.loads(params[0])

        resized_images = []
        for index, image in images.iterrows():
            if not isinstance(image, Image):
                image = Image(**image.to_dict())
            image_pil = image.to_pil()
            scale_factor = params["scaleFactor"]
            if scale_factor != 1.0:
                resized_image = image_pil.resize(
                    (
                        int(image_pil.width * scale_factor),
                        int(image_pil.height * scale_factor),
                    )
                )
            else:
                resized_image = image_pil
            resized_images.append((resized_image, image.path))

        results = cls.call_ocr(resized_images, params)

        return pd.DataFrame(results)

    def outputSchema(self):
        return StructType(
            [
                StructField("path", StringType(), True),
                StructField("text", StringType(), True),
                StructField("type", StringType(), True),
                StructField(
                    "bboxes",
                    ArrayType(
                        StructType(
                            [
                                StructField("text", StringType(), False),
                                StructField("score", DoubleType(), False),
                                StructField("x", IntegerType(), False),
                                StructField("y", IntegerType(), False),
                                StructField("width", IntegerType(), False),
                                StructField("height", IntegerType(), False),
                            ]
                        ),
                        True,
                    ),
                    True,
                ),
                StructField("exception", StringType(), True),
            ]
        )

    def get_params(self):
        return json.dumps({k.name: v for k, v in self.extractParamMap().items()})

    def _transform(self, dataset):
        out_col = self.getOutputCol()
        input_col = self._validate(self.getInputCol(), dataset)
        params = self.get_params()
        if not self.getPartitionMap():
            result = dataset.withColumn(
                out_col,
                udf(self.transform_udf, Document.get_schema())(input_col, lit(params)),
            )
        else:
            if self.getNumPartitions() > 0:
                if self.getPageCol() in dataset.columns:
                    dataset = dataset.repartition(self.getPageCol())
                elif self.getPathCol() in dataset.columns:
                    dataset = dataset.repartition(self.getPathCol())
                dataset = dataset.coalesce(self.getNumPartitions())
            result = dataset.withColumn(
                out_col,
                pandas_udf(self.transform_udf_pandas, self.outputSchema())(input_col, lit(params)),
            )
        if not self.getKeepInputData():
            result = result.drop(input_col)
        return result

    def setLineTolerance(self, value):
        """
        Sets the value of :py:attr:`lineTolerance`.
        """
        return self._set(lineTolerance=value)

    def getLineTolerance(self):
        """
        Gets the value of :py:attr:`lineTolerance`.
        """
        return self.getOrDefault(self.lineTolerance)

    def setKeepFormatting(self, value):
        """
        Sets the value of :py:attr:`keepFormatting`.
        """
        return self._set(keepFormatting=value)

    def getKeepFormatting(self):
        """
        Gets the value of :py:attr:`keepFormatting`.
        """
        return self.getOrDefault(self.keepFormatting)

    def setScaleFactor(self, value):
        """
        Sets the value of :py:attr:`scaleFactor`.
        """
        return self._set(scaleFactor=value)

    def getScaleFactor(self):
        """
        Sets the value of :py:attr:`scaleFactor`.
        """
        return self.getOrDefault(self.scaleFactor)



from pyspark import keyword_only

from scaledp.schemas.Box import Box
from scaledp.schemas.Document import Document
from scaledp.params import *
from .BaseRecognizer import BaseRecognizer
from ...enums import PSM, OEM, TessLib


class TesseractRecognizer(BaseRecognizer):
    """
    Run Tesseract text recognition on images.
    """

    oem = Param(
        Params._dummy(),
        "oem",
        "OCR engine mode. Defaults to :attr:`OEM.DEFAULT`.",
        typeConverter=TypeConverters.toInt,
    )

    tessDataPath = Param(
        Params._dummy(),
        "tessDataPath",
        "Path to tesseract data folder.",
        typeConverter=TypeConverters.toString,
    )

    tessLib = Param(
        Params._dummy(),
        "tessLib",
        "The desired Tesseract library to use. Defaults to :attr:`TESSEROCR`",
        typeConverter=TypeConverters.toInt,
    )

    defaultParams = {
        "inputCols": ["image", "boxes"],
        "outputCol": "text",
        "keepInputData": False,
        "scaleFactor": 1.0,
        "scoreThreshold": 0.5,
        "oem": OEM.DEFAULT,
        "lang": ["eng"],
        "lineTolerance": 0,
        "keepFormatting": False,
        "tessDataPath": "/usr/share/tesseract-ocr/5/tessdata/",
        "tessLib": TessLib.PYTESSERACT,
        "partitionMap": False,
        "numPartitions": 0,
        "pageCol": "page",
        "pathCol": "path",
    }

    @keyword_only
    def __init__(self, **kwargs):
        super(TesseractRecognizer, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)

    @classmethod
    def call_pytesseract(cls, images, boxes, params):
        raise NotImplementedError("Pytesseract version not implemented yet.")
        # import pytesseract
        # results = []
        #
        # config = f"--psm {params['psm']} --oem {params['oem']} -l {cls.getLangTess(params)}"
        # for image, image_path in images:
        #     res = pytesseract.image_to_data(image, output_type=pytesseract.Output.DATAFRAME, config=config)
        #     res["conf"] = res["conf"] / 100
        #
        #     if not params["keepFormatting"]:
        #         res.loc[res["level"] == 4, "conf"] = 1.0
        #         res["text"] = res["text"].fillna('\n')
        #
        #     res = res[res["conf"] > params['scoreThreshold']][['text', 'conf', 'left', 'top', 'width', 'height']] \
        #         .rename(columns={"conf": "score", "left": "x", "top": "y"})
        #     res = res[res["text"] != '\n']
        #     boxes = res.apply(lambda x: Box(*x).toString().scale(1 / params['scaleFactor']), axis=1).values.tolist()
        #     if params["keepFormatting"]:
        #         text = TesseractRecognizer.box_to_formatted_text(boxes, params["lineTolerance"])
        #     else:
        #         text = " ".join([str(w) for w in res["text"].values.tolist()])
        #
        #     results.append(Document(path=image_path,
        #                             text=text,
        #                             type="text",
        #                             bboxes=boxes))
        # return results

    @staticmethod
    def getLangTess(params):
        return "+".join(
            [LANGUAGE_TO_TESSERACT_CODE[CODE_TO_LANGUAGE[lang]] for lang in params["lang"]]
        )

    @classmethod
    def call_tesserocr(cls, images, detected_boxes, params):  # pragma: no cover
        from tesserocr import PyTessBaseAPI

        results = []
        lang = cls.getLangTess(params)
        with PyTessBaseAPI(
            path=params["tessDataPath"],
            psm=PSM.SINGLE_WORD,
            oem=params["oem"],
            lang=lang,
        ) as api:
            api.SetVariable("debug_file", "ocr.log")
            #api.SetVariable("tessedit_char_whitelist", "1234567890.,-:")

            for (image, image_path), detected_box in zip(images, detected_boxes):
                api.SetImage(image)

                boxes = []
                texts = []
                # if not isinstance(detected_box, DetectorOutput):
                #     detected_box = DetectorOutput(**detected_box.to_dict())

                for box in detected_box.bboxes:
                    if isinstance(box, dict):
                        box = Box(**box)
                    if not isinstance(box, Box):
                        box = Box(**box.asDict())
                    scaled_box = box.scale(params["scaleFactor"], padding=0)
                    api.SetRectangle(
                        scaled_box.x, scaled_box.y, scaled_box.width, scaled_box.height
                    )
                    box.text = api.GetUTF8Text().replace("\n", "")
                    box.score = api.MeanTextConf() / 100
                    if box.score > params["scoreThreshold"]:
                        boxes.append(box)
                        texts.append(box.text)
                if params["keepFormatting"]:
                    text = TesseractRecognizer.box_to_formatted_text(
                        boxes, params["lineTolerance"]
                    )
                else:
                    text = " ".join(texts)

                results.append(
                    Document(
                        path=image_path,
                        text=text,
                        bboxes=boxes,
                        type="text",
                        exception="",
                    )
                )
        return results

    @classmethod
    def call_recognizer(cls, images, boxes, params):
        if params["tessLib"] == TessLib.TESSEROCR.value:
            return cls.call_tesserocr(images, boxes, params)
        elif params["tessLib"] == TessLib.PYTESSERACT.value:
            return cls.call_pytesseract(images, boxes, params)
        else:
            raise ValueError(f"Unknown Tesseract library: {params['tessLib']}")

    def setOem(self, value):
        """
        Sets the value of :py:attr:`oem`.
        """
        return self._set(oem=value)

    def getOem(self):
        """
        Sets the value of :py:attr:`oem`.
        """
        return self.getOrDefault(self.oem)

    def setTessDataPath(self, value):
        """
        Sets the value of :py:attr:`tessDataPath`.
        """
        return self._set(tessDataPath=value)

    def getTessDataPath(self):
        """
        Sets the value of :py:attr:`tessDataPath`.
        """
        return self.getOrDefault(self.tessDataPath)

    def setTessLib(self, value):
        """
        Sets the value of :py:attr:`tessLib`.
        """
        return self._set(tessLib=value)

    def getTessLib(self):
        """
        Gets the value of :py:attr:`tessLib`.
        """
        return self.getOrDefault(self.tessLib)



# Description: Schema Detector Output
from scaledp.utils.dataclass import map_dataclass_to_struct, register_type
from dataclasses import dataclass
from scaledp.schemas.Box import Box


@dataclass(order=True)
class DetectorOutput:
    path: str
    type: str
    bboxes: list[Box]
    exception: str = ""

    @staticmethod
    def get_schema():
        return map_dataclass_to_struct(DetectorOutput)


register_type(DetectorOutput, DetectorOutput.get_schema)



from pyspark.sql.types import *
import imagesize
import io
import logging
from PIL import Image as pImage
import traceback
from dataclasses import dataclass
from ..enums import ImageType
from scaledp.utils.dataclass import map_dataclass_to_struct, register_type, BinaryT


@dataclass(order=True)
class Image(object):
    """
    Image object for represent image data in Spark Dataframe
    """

    path: str
    resolution: int = 0
    data: BinaryT = bytes()
    imageType: str = ImageType.FILE.value
    exception: str = ""
    height: int = 0
    width: int = 0

    def to_pil(self):
        if self.imageType == ImageType.FILE.value or self.imageType == ImageType.WEBP.value:
            return pImage.open(io.BytesIO(self.data))

    def to_io_stream(self):
        return io.BytesIO(self.data)

    def to_opencv(self):
        if self.imageType == ImageType.FILE.value:
            return pImage.open(io.BytesIO(self.data))

    def to_webp(self):
        if self.imageType == ImageType.FILE.value:
            image = pImage.open(io.BytesIO(self.data))
            buff = io.BytesIO()
            image.save(buff, "webp")
            self.data = buff.getvalue()
        return self

    @staticmethod
    def from_binary(data, path, imageType, resolution=None, width=None, height=None):
        img = Image(path=path, data=data, imageType=ImageType.FILE.value, resolution=resolution)
        if data is None or len(data) == 0:
            raise ValueError("Empty image data.")
        if imageType in (ImageType.FILE.value, ImageType.WEBP.value):
            if height is not None:
                img.height = height
            if width is not None:
                img.width = width
            if width is None and height is None:
                img.width, img.height = imagesize.get(io.BytesIO(img.data))
                if img.width == -1:
                    raise Exception("Unable to read image.")
                logging.info(f"Image size: {img.width}x{img.height}")
        return img

    @staticmethod
    def from_pil(data, path, imageType, resolution):
        buff = io.BytesIO()
        if imageType == ImageType.WEBP.value:
            data.save(buff, "webp")
        else:
            data.save(buff, "png")
        img = Image(
            path=path,
            data=buff.getvalue(),
            imageType=ImageType.FILE.value,
            width=data.width,
            height=data.height,
            resolution=resolution,
        )
        return img

    @staticmethod
    def get_schema():
        return map_dataclass_to_struct(Image)

    def __str__(self):
        return f"Image(path={self.path}, resolution={self.resolution}, imageType={self.imageType}, exception={self.exception}, height={self.height}, width={self.width})"

    def __repr__(self):
        return self.__str__()


register_type(Image, Image.get_schema)



import json
import time
import logging

from .BaseExtractor import BaseExtractor
from pyspark import keyword_only

from ...params import HasLLM, HasSchema, HasPrompt
from ...schemas.ExtractorOutput import ExtractorOutput
from tenacity import (
    retry,
    stop_after_attempt,
    wait_random_exponential,
    retry_if_exception_type
)


class LLMExtractor(BaseExtractor, HasLLM, HasSchema, HasPrompt):

    defaultParams = {
        "inputCol": "text",
        "outputCol": "data",
        "keepInputData": True,
        "model": "llama3-8b-8192",
        "apiBase": None,
        "apiKey": None,
        "numPartitions": 1,
        "pageCol": "page",
        "pathCol": "path",
        "prompt": """Please extract data from the text as json.""",
        "systemPrompt": "You are data extractor from the scanned images.",
        "delay": 30,
        "maxRetry": 6,
        "propagateError": False,
        "temperature": 1.0,
        "schemaByPrompt": True
    }

    @keyword_only
    def __init__(self, **kwargs):
        super(LLMExtractor, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)
        self.pipeline = None

    def call_extractor(self, documents, params):
        from openai import RateLimitError
        client = self.getOIClient()

        @retry(retry=retry_if_exception_type(RateLimitError),
               wait=wait_random_exponential(min=1,
                                            max=self.getDelay()),
                                            stop=stop_after_attempt(self.getMaxRetry()))
        def completion_with_backoff(**kwargs):
            logging.info(f"Calling LLM API")
            return client.beta.chat.completions.parse(**kwargs)

        results = []
        for document in documents:
            data = document.text

            content = [
                {
                    "type": "text",
                    "text": params["prompt"]
                }
            ]
            kwargs = {}

            if self.getSchemaByPrompt():
                content.append({
                    "type": "text",
                    "text": "Schema for the output json: " + self.getSchema() + " Always return valid json. Do not include schema to the output."
                })
                # kwargs["response_format"] = {"type": "json_object"}
            else:
                kwargs["response_format"] = self.getPaydanticSchema()

            completion = completion_with_backoff(
                model=params["model"],
                messages=[
                    {
                        "role": "system",
                        "content": params["systemPrompt"],
                        "role": "user",
                        "content":
                            content + [
                                {
                                    "type": "text",
                                    "text": data
                                },
                            ]
                    } ,
                ],
                response_format=self.getPaydanticSchema(),
                **kwargs
            )
            result = completion.choices[0].message.content.replace("```json", "").replace("```", "")

            results.append(
                ExtractorOutput(
                    path=document.path,
                    #data=json.dumps(completion.choices[0].message.parsed.json(),indent=4, ensure_ascii=False),
                    #completion.choices[0].message.parsed.json()
                    data=json.dumps(result,indent=4, ensure_ascii=False),
                    type="LLMExtractor",
                    exception="",
                )
            )
        return results



import traceback
import logging

from pyspark import keyword_only
from pyspark.sql.functions import udf
from pyspark.ml import Transformer
from pyspark.ml.util import DefaultParamsReadable, DefaultParamsWritable

from scaledp.schemas.Box import Box
from scaledp.schemas.Image import Image
from ..enums import ImageType
from scaledp.params import *


class ImageCropError(Exception):
    pass


class ImageCropBoxes(
    Transformer,
    HasInputCols,
    HasOutputCol,
    HasKeepInputData,
    HasImageType,
    HasPageCol,
    DefaultParamsReadable,
    DefaultParamsWritable,
    HasColor,
    HasNumPartitions,
    HasColumnValidator,
    HasDefaultEnum,
    HasPropagateError,
    metaclass=AutoParamsMeta,
):
    """
    Crop image by bounding boxes
    """

    padding = Param(Params._dummy(), "padding", "Padding.", typeConverter=TypeConverters.toInt)
    noCrop = Param(Params._dummy(), "noCrop", "Does not Crop if boxes is empty.", typeConverter=TypeConverters.toBoolean)

    defaultParams = {
        "inputCols": ["image", "boxes"],
        "outputCol": "cropped_image",
        "keepInputData": False,
        "imageType": ImageType.FILE,
        "numPartitions": 0,
        "padding": 0,
        "pageCol": "page",
        "propagateError": False,
        "noCrop": True,
    }

    @keyword_only
    def __init__(self, **kwargs):
        super(ImageCropBoxes, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)

    def transform_udf(self, image, data):
        if not isinstance(image, Image):
            image = Image(**image.asDict())
        try:
            if image.exception != "":
                return Image(
                    path=image.path,
                    imageType=image.imageType,
                    data=bytes(),
                    exception=image.exception,
                )
            img = image.to_pil()
            results = []
            for box in data.bboxes:
                if not isinstance(box, Box):
                    box = Box(**box.asDict())
                if box.width > box.height:
                    results.append(img.crop(box.bbox(self.getPadding())).rotate(-90, expand=True))
                else:
                    results.append(img.crop(box.bbox(self.getPadding())))
            if self.getNoCrop() and len(results) == 0:
                raise ImageCropError("No boxes to crop")
            if len(results) == 0:
                results.append(img)

        except Exception as e:
            exception = traceback.format_exc()
            exception = f"ImageCropBoxes: {exception}, {image.exception}"
            logging.warning(exception)
            if self.getPropagateError():
                raise ImageCropError() from e
            return Image(image.path, image.imageType, data=bytes(), exception=exception)
        return Image.from_pil(results[0], image.path, image.imageType, image.resolution)

    def _transform(self, dataset):
        out_col = self.getOutputCol()
        image_col = self._validate(self.getInputCols()[0], dataset)
        box_col = self._validate(self.getInputCols()[1], dataset)

        if self.getNumPartitions() > 0:
            dataset = dataset.repartition(self.getPageCol()).coalesce(self.getNumPartitions())
        result = dataset.withColumn(
            out_col, udf(self.transform_udf, Image.get_schema())(image_col, box_col)
        )

        if not self.getKeepInputData():
            result = result.drop(image_col)
        return result



from pyspark import keyword_only

from ...enums import Device
from scaledp.models.detectors.BaseDetector import BaseDetector
from scaledp.params import HasDevice, HasBatchSize
import gc
import io
from typing import Any

from scaledp.schemas.Box import Box
from scaledp.schemas.DetectorOutput import DetectorOutput


class DocTRTextDetector(BaseDetector, HasDevice, HasBatchSize):

    defaultParams = {
        "inputCol": "image",
        "outputCol": "boxes",
        "keepInputData": False,
        "scaleFactor": 1.0,
        "scoreThreshold": 0.1,
        "device": Device.CPU,
        "batchSize": 2,
        "model": "db_resnet50",
        "partitionMap": False,
        "numPartitions": 0,
        "pageCol": "page",
        "pathCol": "path",
    }

    @keyword_only
    def __init__(self, **kwargs):
        super(DocTRTextDetector, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)

    @classmethod
    def call_detector(cls, images, params):
        from doctr.io import DocumentFile
        from doctr.models import ocr_predictor
        from doctr.file_utils import CLASS_NAME
        import torch

        predictor = ocr_predictor(pretrained=True, det_arch=params["model"])

        predictor.det_predictor.model.postprocessor.box_thresh = params["scoreThreshold"]

        detector = predictor.det_predictor

        if int(params["device"]) == Device.CPU.value:
            device = "cpu"
        else:
            device = "cuda"

        def resolve_geometry(
            geom: Any,
        ) -> (
            tuple[float, float, float, float]
            | tuple[float, float, float, float, float, float, float, float]
        ):
            if len(geom) == 4:
                return (*geom[0], *geom[1], *geom[2], *geom[3])
            return (*geom[0], *geom[1])

        docs = []
        for image, image_path in images:
            buff = io.BytesIO()
            image.save(buff, "png")
            docs.extend(DocumentFile.from_images(buff.getvalue()))

        results = detector.to(device)(docs)

        results_final = []
        for result, (image, image_path) in zip(results, images):
            boxes = []
            w, h = image.size
            for geom in result[CLASS_NAME]:
                g = (
                    geom[:-1].tolist()
                    if geom.shape == (5,)
                    else resolve_geometry(geom[:4].tolist())
                )
                boxes.append(
                    Box.fromBBox([g[0] * w, g[1] * h, g[2] * w, g[3] * h], score=geom[-1])
                )
            results_final.append(DetectorOutput(path=image_path, type="doctr", bboxes=boxes))

        gc.collect()
        if int(params["device"]) == Device.CUDA.value:
            torch.cuda.empty_cache()

        return results_final



import datetime
import logging
from typing import List

from pydantic import BaseModel
from pyspark import keyword_only
from pyspark.ml.param import Param, Params, TypeConverters

from scaledp.params import HasLLM, HasPrompt, HasPropagateError
from scaledp.models.ner.BaseNer import BaseNer
from ...enums import Device
from scaledp.schemas.Entity import Entity
from scaledp.schemas.NerOutput import NerOutput
import pandas as pd
import json

from tenacity import (
    retry,
    stop_after_attempt,
    wait_random_exponential,
    retry_if_exception_type
)


def log_info(msg):
    print(f"{datetime.datetime.now()} INFO: {msg}")


class LLMNer(BaseNer, HasLLM, HasPrompt, HasPropagateError):

    tags = Param(
        Params._dummy(),
        "tags",
        "Ner tags.",
        typeConverter=TypeConverters.toListString,
    )

    defaultParams = {
        "inputCol": "text",
        "outputCol": "ner",
        "keepInputData": True,
        "model": "d4data/biomedical-ner-all",
        "whiteList": [],
        "numPartitions": 1,
        "device": Device.CPU,
        "batchSize": 1,
        "scoreThreshold": 0.0,
        "pageCol": "page",
        "pathCol": "path",
        "systemPrompt": "Please extract following NER Tags from the text as json: ",
        "prompt": """Please extract text from the image.""",
        "model": "gemini-1.5-flash-8b",
        "apiBase": "",
        "apiKey": "",
        "delay": 30,
        "maxRetry": 6,
        "propagateError": False,
        "tags": ["PERSON", "LOCATION", "DATE", "EMAIL", "PHONE", "ORGANIZATION"],
        "partitionMap": False,
    }

    def getPaydanticSchema(self):

        class Entity(BaseModel):
            entity_group: str
            score: float
            word: str
            start: int
            end: int

        class NerOutput(BaseModel):
            entities: List[Entity]

        return NerOutput

    @keyword_only
    def __init__(self, **kwargs):
        super(LLMNer, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)
        self.pipeline = None

    def transform_udf(self, document, params=None):

        if params is None:
            params = self.get_params()
        params = json.loads(params)

        mapping = []
        for idx, box in enumerate(document.bboxes):
            mapping.extend([idx] * (len(box.text) + 1))

        from openai import RateLimitError

        client = self.getClient(params["apiKey"], params["apiBase"])
        results = []

        @retry(retry=retry_if_exception_type(RateLimitError),
               wait=wait_random_exponential(min=1,
                                            max=params["delay"]),
               stop=stop_after_attempt(params["maxRetry"]))
        def completion_with_backoff(**kwargs):
            logging.info(f"Calling LLM API")
            return client.beta.chat.completions.parse(**kwargs)

        schema = self.getPaydanticSchema().model_json_schema()

        completion = completion_with_backoff(
            model=params["model"],
            messages=[
                {
                    #"role": "system",
                    #"content": "You are NER",#params["systemPrompt"] + ",".join(params["tags"]),
                    "role": "user",
                    "content": f'Pleas extract NER tags: {",".join(params["tags"])} as json with schema: {schema}. From the text:' + document.text
                },
            ],
            response_format={"type": "json_object"} #self.getPaydanticSchema(),
        )
        #
        entities = []
        #data = json.loads(completion.choices[0].message.parsed.json())
        data = json.loads(completion.choices[0].message.content.replace("```json", "").replace("```", ""))
        for tag in data["entities"]:
            if len(self.getWhiteList()) > 0 and tag["entity_group"] not in self.getWhiteList():
                continue
            # boxes = mapping[tag["start"] : tag["end"]]
            # boxes = [document.bboxes[i] for i in list(dict.fromkeys(boxes))]
            boxes = []
            word = tag["word"]
            for idx, box in enumerate(document.bboxes):
                if any(word.lower() in box.text.lower() for word in word.split(" ") if len(word) > 2):
                    boxes.append(box)

            t = Entity(
                entity_group=tag["entity_group"],
                score=float(tag["score"]),
                word=tag["word"],
                start=tag["start"],
                end=tag["end"],
                boxes=boxes,
            )
            entities.append(t)

        output = NerOutput(path=document.path, entities=entities, exception="")
        return output

    @staticmethod
    def transform_udf_pandas(
        texts: pd.DataFrame, params: pd.Series
    ) -> pd.DataFrame:  # pragma: no cover
        params = json.loads(params[0])
        model = params["model"]

        results = []

        return pd.DataFrame(results)



import traceback
import logging
import random

from PIL import ImageDraw
from pyspark import keyword_only
from pyspark.sql.functions import udf
from pyspark.ml import Transformer
from pyspark.ml.util import DefaultParamsReadable, DefaultParamsWritable

from scaledp.schemas.Box import Box
from scaledp.schemas.Entity import Entity
from scaledp.schemas.Image import Image
from scaledp.schemas.NerOutput import NerOutput
from ..enums import ImageType
from scaledp.params import *


class ImageDrawBoxes(
    Transformer,
    HasInputCols,
    HasOutputCol,
    HasKeepInputData,
    HasImageType,
    HasPageCol,
    DefaultParamsReadable,
    DefaultParamsWritable,
    HasColor,
    HasNumPartitions,
    HasColumnValidator,
    HasDefaultEnum,
    HasWhiteList,
    HasBlackList,
    metaclass=AutoParamsMeta,
):
    """
    Draw boxes on image
    """

    filled = Param(
        Params._dummy(),
        "filled",
        "Fill rectangle.",
        typeConverter=TypeConverters.toBoolean,
    )

    lineWidth = Param(
        Params._dummy(), "lineWidth", "Line width.", typeConverter=TypeConverters.toInt
    )

    textSize = Param(Params._dummy(), "textSize", "Text size.", typeConverter=TypeConverters.toInt)

    padding = Param(Params._dummy(), "padding", "Padding.", typeConverter=TypeConverters.toInt)

    displayDataList = Param(
        Params._dummy(),
        "displayDataList",
        "Display data list.",
        typeConverter=TypeConverters.toListString,
    )

    defaultParams = {
        "inputCols": ["image", "boxes"],
        "outputCol": "image_with_boxes",
        "keepInputData": False,
        "imageType": ImageType.FILE,
        "filled": False,
        "color": None,
        "lineWidth": 1,
        "textSize": 12,
        "displayDataList": [],
        "numPartitions": 0,
        "padding": 0,
        "pageCol": "page",
        "whiteList": [],
        "blackList": [],
    }

    @keyword_only
    def __init__(self, **kwargs):
        super(ImageDrawBoxes, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)

    def getDisplayText(self, box):
        display_list = self.getDisplayDataList()
        text = []
        if display_list:
            for name in display_list:
                if hasattr(box, name):
                    val = getattr(box, name)
                    if isinstance(val, float):
                        text.append(f"{val:0.2f}")
                    elif isinstance(val, int):
                        text.append(str(val))
                    else:
                        text.append(val)
        return ":".join(text)

    def transform_udf(self, image, data):

        def get_color():
            return "#{:06x}".format(random.randint(0, 0xFFFFFF))

        if not isinstance(image, Image):
            image = Image(**image.asDict())
        try:
            if image.exception != "":
                return Image(
                    image.origin,
                    image.imageType,
                    data=bytes(),
                    exception=image.exception,
                )
            img = image.to_pil()
            img1 = ImageDraw.Draw(img)
            fill = self.getColor() if self.getFilled() else None

            black_list = self.getBlackList()
            white_list = self.getWhiteList()

            if hasattr(data, "entities"):
                if not isinstance(data, NerOutput):
                    data = NerOutput(**data.asDict())
                colors = {}
                for ner in data.entities:

                    if not isinstance(ner, Entity):
                        ner = Entity(**ner.asDict())

                    if ner.entity_group in black_list:
                        continue
                    if white_list and ner.entity_group not in white_list:
                        continue

                    if ner.entity_group not in colors:
                        colors[ner.entity_group] = get_color()
                    if self.getColor() is None:
                        color = colors[ner.entity_group]
                    else:
                        color = self.getColor()
                    for box in ner.boxes:
                        if not isinstance(box, Box):
                            box = Box(**box.asDict())
                        text = self.getDisplayText(ner)
                        img1.rounded_rectangle(
                            box.shape(self.getPadding()),
                            outline=color,
                            radius=4,
                            fill=fill,
                            width=self.getLineWidth(),
                        )
                        if text:
                            tbox = list(
                                img1.textbbox(
                                    (
                                        box.x,
                                        box.y - self.getTextSize() * 1.2 - self.getPadding(),
                                    ),
                                    text,
                                    font_size=self.getTextSize(),
                                )
                            )
                            tbox[3] = tbox[3] + self.getTextSize() / 4
                            tbox[2] = tbox[2] + self.getTextSize() / 4
                            tbox[0] = box.x - self.getPadding()
                            tbox[1] = box.y - self.getTextSize() * 1.2 - self.getPadding()
                            img1.rounded_rectangle(tbox, outline=color, radius=2, fill=color)
                            img1.text(
                                (
                                    box.x,
                                    box.y - self.getTextSize() * 1.2 - self.getPadding(),
                                ),
                                text,
                                stroke_width=0,
                                fill="white",
                                font_size=self.getTextSize(),
                            )
            else:
                if self.getColor() is None:
                    color = "green"
                else:
                    color = self.getColor()
                for box in data.bboxes:
                    if not isinstance(box, Box):
                        box = Box(**box.asDict())
                    img1.rounded_rectangle(
                        box.shape(self.getPadding()),
                        outline=color,
                        radius=4,
                        fill=fill,
                        width=self.getLineWidth(),
                    )
                    text = self.getDisplayText(box)
                    if text:
                        img1.text(
                            (
                                box.x,
                                box.y - self.getTextSize() * 1.2 - self.getPadding(),
                            ),
                            text,
                            fill=color,
                            font_size=self.getTextSize(),
                        )

        except Exception:
            exception = traceback.format_exc()
            exception = f"ImageDrawBoxes: {exception}, {image.exception}"
            logging.warning(exception)
            return Image(image.path, image.imageType, data=bytes(), exception=exception)
        return Image.from_pil(img, image.path, image.imageType, image.resolution)

    def _preprocessing(self, dataset):
        return dataset

    def _transform(self, dataset):
        out_col = self.getOutputCol()
        image_col = self._validate(self.getInputCols()[0], dataset)
        box_col = self._validate(self.getInputCols()[1], dataset)

        dataset = self._preprocessing(dataset)

        if self.getNumPartitions() > 0:
            dataset = dataset.repartition(self.getPageCol()).coalesce(self.getNumPartitions())
        result = dataset.withColumn(
            out_col, udf(self.transform_udf, Image.get_schema())(image_col, box_col)
        )

        if not self.getKeepInputData():
            result = result.drop(image_col)
        return result



## Form the https://github.com/xdanny/pyspark_types/blob/main/pyspark_types/auxiliary.py
## with some changes

from typing import Type
from decimal import Decimal


class LongT(int):
    def __repr__(self):
        return f"LongT({super().__repr__()})"


class ShortT(int):
    def __repr__(self):
        return f"ShortT({super().__repr__()})"


class ByteT(int):
    def __repr__(self):
        return f"ByteT({super().__repr__()})"


class BinaryT(int):
    def __repr__(self):
        return f"BinaryT({super().__repr__()})"


class BoundDecimal(Decimal):
    """
    Custom data type that represents a decimal with a specific scale and precision.
    """

    def __new__(cls, value: str, precision: int, scale: int):
        obj = super().__new__(cls, value)
        obj.precision = precision
        obj.scale = scale
        return obj

    def __repr__(self) -> str:
        return f"BoundDecimal('{str(self)}', precision={self.precision}, scale={self.scale})"


def create_bound_decimal_type(precision: int, scale: int) -> Type[BoundDecimal]:
    """
    Factory method that creates a new BoundDecimal type with the specified precision and scale.
    """

    class _BoundDecimal(BoundDecimal):
        def __new__(cls, value):
            return super().__new__(cls, value, precision=precision, scale=scale)

    _BoundDecimal.__name__ = f"BoundDecimal_{precision}_{scale}"
    _BoundDecimal.precision = precision
    _BoundDecimal.scale = scale

    return _BoundDecimal



from dataclasses import dataclass
from scaledp.utils.dataclass import (
    map_dataclass_to_struct,
    register_type,
    apply_nullability,
)
from scaledp.schemas.Box import Box


@dataclass(order=True)
class Entity:
    entity_group: str
    score: float
    word: str
    start: int
    end: int
    boxes: list[Box]

    @staticmethod
    def get_schema():
        return apply_nullability(map_dataclass_to_struct(Entity), True)


register_type(Entity, Entity.get_schema)



from pyspark.sql.functions import input_file_name, udf
from pyspark.ml import Transformer
from pyspark.ml.util import DefaultParamsReadable, DefaultParamsWritable
from scaledp.params import *
from scaledp.schemas.Document import Document


class TextToDocument(
    Transformer, HasInputCol, HasOutputCol, DefaultParamsReadable, DefaultParamsWritable
):

    def __init__(self, inputCol="value", outputCol="text"):
        super(TextToDocument, self).__init__()
        self._setDefault(inputCol=inputCol, outputCol=outputCol)

    @staticmethod
    def transform_udf(text, path):
        return Document(
            path=path,
            text=text,
            type="text",
            bboxes=[],
        )

    def _transform(self, dataset):
        input_col = self.getInputCol()
        output_col = self.getOutputCol()

        text_to_document_udf = udf(self.transform_udf, Document.get_schema())
        return dataset.withColumn(
            output_col, text_to_document_udf(dataset[input_col], input_file_name())
        )



from enum import Enum
from pydantic import BaseModel, Field, create_model
from typing import Any, Dict, List, Optional, Type, Union
from datetime import datetime, date, time

def json_schema_to_model(json_schema: Dict[str, Any], definitions: Dict[str, Any] = None) -> Type[BaseModel]:
    """
    Converts a JSON schema to a Pydantic BaseModel class.

    Args:
        json_schema: The JSON schema to convert.
        definitions: A dictionary of defined schemas, often found in the "$defs" property.

    Returns:
        A Pydantic BaseModel class.
    """

    model_name = json_schema.get('title')
    field_definitions = {
        name: json_schema_to_pydantic_field(name, prop, json_schema.get('required', []), definitions)
        for name, prop in json_schema.get('properties', {}).items()
    }
    return create_model(model_name, **field_definitions)

def json_schema_to_pydantic_field(name: str, json_schema: Dict[str, Any], required: List[str], definitions: Dict[str, Any] = None) -> Any:
    """
    Converts a JSON schema property to a Pydantic field definition.

    Args:
        name: The field name.
        json_schema: The JSON schema property.
        required: A list of required field names.
        definitions: A dictionary of defined schemas, often found in the "$defs" property.

    Returns:
        A Pydantic field definition.
    """

    type_ = json_schema_to_pydantic_type(json_schema, definitions)
    description = json_schema.get('description')
    examples = json_schema.get('examples')
    min_length = json_schema.get('minLength')
    max_length = json_schema.get('maxLength')
    minimum = json_schema.get('minimum')
    maximum = json_schema.get('maximum')
    exclusive_minimum = json_schema.get('exclusiveMinimum')
    exclusive_maximum = json_schema.get('exclusiveMaximum')
    pattern = json_schema.get('pattern')

    # Handle optional fields (including 'anyOf' with 'null')
    if name not in required or ('anyOf' in json_schema and any(item.get('type') == 'null' for item in json_schema.get('anyOf', []))):
        type_ = Optional[type_]

    return (type_, Field(description=description,
                         examples=examples,
                         min_length=min_length,
                         max_length=max_length,
                         ge=minimum,
                         le=maximum,
                         gt=exclusive_minimum,
                         lt=exclusive_maximum,
                         pattern=pattern,
                         default=... if name in required else None))

def json_schema_to_pydantic_type(json_schema: Dict[str, Any], definitions: Dict[str, Any] = None) -> Any:
    """
    Converts a JSON schema type to a Pydantic type.

    Args:
        json_schema: The JSON schema to convert.
        definitions: A dictionary of defined schemas, often found in the "$defs" property.

    Returns:
        A Pydantic type.
    """

    type_ = json_schema.get('type')

    if type_ == 'string':
        format_ = json_schema.get('format')
        if format_ == 'date':
            return date
        elif format_ == 'time':
            return time
        elif format_ == 'date-time':
            return datetime
        if 'enum' in json_schema:
            return Enum(json_schema.get('title', 'EnumType'), {member: member for member in json_schema['enum']})
        return str
    elif type_ == 'integer':
        return int
    elif type_ == 'number':
        return float
    elif type_ == 'boolean':
        return bool
    elif type_ == 'array':
        items_schema = json_schema.get('items')
        if items_schema:
            item_type = json_schema_to_pydantic_type(items_schema, definitions)
            return List[item_type]
        else:
            return List
    elif type_ == 'object':
        # Handle nested models.
        properties = json_schema.get('properties')
        if properties:
            nested_model = json_schema_to_model(json_schema, definitions)
            return nested_model
        else:
            return Dict
    elif type_ == 'null':
        return None
    elif '$ref' in json_schema:
        # Handle references to nested schemas
        ref_path = json_schema['$ref'].split('/')
        ref_name = ref_path[-1]
        if definitions:
            ref_schema = definitions.get(ref_name)
            if ref_schema:
                return json_schema_to_pydantic_type(ref_schema, definitions)
        raise ValueError(f"Could not resolve reference: {json_schema['$ref']}")
    elif 'anyOf' in json_schema:
        # Handle 'anyOf' with 'null' for optional fields
        types = [json_schema_to_pydantic_type(item, definitions) for item in json_schema['anyOf']]
        if any(t is None for t in types):  # If 'null' is present in anyOf
            types = [t for t in types if t is not None]  # Remove 'null' from the list
            if len(types) == 1:
                return Optional[types[0]]  # If only one non-null type, use Optional
            else:
                return Union[types]  # If multiple non-null types, use Union
        else:
            # If 'null' is not present, return the union of all types
            return Union[types]
    else:
        raise ValueError(f'Unsupported JSON schema type: {type_}')



from pyspark import keyword_only
from pyspark.sql.functions import udf, lit
from pyspark.ml import Transformer
from pyspark.ml.util import DefaultParamsReadable, DefaultParamsWritable
from scaledp.schemas.Image import Image
from scaledp.params import *
from scaledp.enums import ImageType
import logging
import traceback


class ImageError(Exception):
    pass


class DataToImage(
    Transformer,
    HasInputCol,
    HasOutputCol,
    HasKeepInputData,
    HasImageType,
    HasDefaultEnum,
    HasPathCol,
    DefaultParamsReadable,
    DefaultParamsWritable,
    HasColumnValidator,
    HasPropagateError,
):
    """
    Transform Binary Content to Image
    """

    defaultParams = {
        "inputCol": "content",
        "outputCol": "image",
        "pathCol": "path",
        "keepInputData": False,
        "imageType": ImageType.FILE,
        "propagateError": False,
    }

    @keyword_only
    def __init__(self, **kwargs):
        super(DataToImage, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)

    def transform_udf(self, input, path, resolution):
        try:
            return Image.from_binary(input, path, self.getImageType(), resolution=resolution)
        except Exception as ex:
            exception = traceback.format_exc()
            exception = f"DataToImage: {exception}"
            logging.warning(exception)
            if self.getPropagateError():
                raise ImageError(exception) from ex
            return Image(path, self.getImageType(), data=bytes(), exception=exception)


    def _transform(self, dataset):
        out_col = self.getOutputCol()
        input_col = self._validate(self.getInputCol(), dataset)
        path_col = self._validate(self.getPathCol(), dataset)
        if "resolution" in dataset.columns:
            resolution = dataset["resolution"]
        else:
            resolution = lit(0)
        result = dataset.withColumn(
            out_col,
            udf(self.transform_udf, Image.get_schema())(input_col, path_col, resolution),
        )
        if not self.getKeepInputData():
            result = result.drop(input_col)
        return result
